#include "msl.hpp"

namespace v = msl::var;

void
set_x_100(msl::Ref ref)
{
	ref.x = 100;
}
void
set_x_0(msl::Ref ref)
{
	ref.x = 0;
}

int
main()
{
	auto shuffle{ true };
	size_t const size{ 10 * msl::THREADS_NR };
	std::vector<size_t> indices(size);
	std::iota(indices.begin(), indices.end(), 0);
	std::mt19937 gnr{};
	if (shuffle)
		std::shuffle(indices.begin(), indices.end(), gnr);

	{
		msl::Population popul(100'000);
		for (auto [i, ref] : msl::enumerate(popul))
			ref.set_id_number(i)
			  .set_alive(i % 10)
			  .set_female(gnr() % 2)
			  .set_birthday_ymd(2025 - 100 + gnr() % 100, 1, 1)
			  .set_family_head(true)
			  .set_family_next(ref)
			  .set_household_head(true)
			  .set_household_next(ref)
			  .set_tax_fam_head(true)
			  .set_tax_fam_next(ref)
			  .set_ben_fam_head(true)
			  .set_ben_fam_next(ref);

		msl::to_day = msl::days_from_civil(2025, 1, 1);
		MSL_SHOW(popul.avg(MSL_LMBD(ref.age())))

		// for (int t{ 2026 }; t < 2026 + 100; ++t) {
		for (int t{ 2026 }; t < 2026 + 2; ++t) {
			// Set the date to January 1st of the current year.
			msl::to_day = msl::days_from_civil(t, 1, 1);
			// The births.
			popul.fill(v::prob, 0);
			// Uniform probability for each female aged from 20 to 40, 2 childs per woman.
			popul.fill_if(v::prob, 2. / 20, MSL_LMBD(ref.female() && msl::within(20, ref.age(), 40)));
			// Alignement on the sum of individual probabilities.
			auto const births{ size_t(popul.sum(v::prob) + .5) };
			// Simulate the events, set the sex of the baby.
			popul.align(births, v::prob, [&](msl::Ref mother) {
				msl::Ref baby{ popul.birth(mother, msl::to_day - gnr() % 365) };
				baby.set_female(gnr() % 2);
			});
			// The deaths.
			popul.fill(v::prob, 0);
			// Uniform probability for each individual over 60.
			popul.fill_if(v::prob, .05, MSL_LMBD(ref.age() > 60));
			// Alignement on the sum of individual probabilities.
			auto const deaths{ size_t(popul.sum(v::prob) + .5) };
			// Simulate the events.
			popul.align(deaths, v::prob, MSL_FNCTN(ref.death()));
			// Some statistics.
			auto const pop_sz{ popul.count_if(v::one) };
			auto const avg_age{ popul.avg(MSL_LMBD(ref.age())) };
			std::cout << t << ' ' << pop_sz << ' ' << size_t(avg_age + .5) << ' ' << births << ' '
			          << deaths << '\n';
		}
	}
	{
		std::cout << "Function member Population::count_if\n";
		msl::Population popul(size);
		for (auto [i, ii] : msl::enumerate(indices))
			popul[ii].set_alive(i % 10).set_std_living_10((i + 1) % 10).set_x(i % 3);

		// On a FLG data member.
		{
			size_t count{};
			for (msl::RefConst ref : popul)
				count += ref.alive();
			assert(popul.count_if(v::one) == count);
		}

		// On a CTG data member.
		{
			size_t count{};
			for (msl::RefConst ref : popul)
				count += ref.alive() && ref.std_living_10;
			assert(popul.count_if(v::std_living_10) == count);
		}

		// On a FLT data member.
		{
			size_t count{};
			for (auto [i, ref] : msl::enumerate(popul))
				count += ref.alive() && ref.x;
			assert(popul.count_if(v::x) == count);
		}

		// Throw an error : invalid variable.
		try {
			popul.count_if(v::family_next);
			assert("not reached");
		} catch (...) {
		}
		// Exit on error : invalid data.
		// popul.fill_with_missing(v::x);
		// popul.count_if(v::x);

		// Using a lambda.
		assert(popul.count_if(MSL_LMBD(true)) == popul.count_if(v::one));
	}
	{
		std::cout << "Function member Population::sum\n";
		msl::Population popul(size);
		for (auto [i, ii] : msl::enumerate(indices))
			popul[ii].set_alive(i % 10).set_std_living_10((i + 1) % 10).set_x(i % 3);

		// On a FLG data member.
		{
			size_t count{};
			for (msl::RefConst ref : popul)
				count += ref.alive();
			assert(popul.sum(v::one) == count);
		}

		// On a CTG data member.
		{
			size_t count{};
			for (msl::RefConst ref : popul)
				count += ref.alive() ? ref.std_living_10 : 0;
			assert(popul.sum(v::std_living_10) == count);
		}

		// On a FLT data member.
		{
			size_t count{};
			for (auto [i, ref] : msl::enumerate(popul))
				ref.x = i % 2, count += ref.alive() ? i % 2 : 0;
			assert(popul.sum(v::x) == count);
		}

		// Throw an error : invalid variable.
		try {
			popul.sum(v::family_next);
			assert("not reached");
		} catch (...) {
		}
		// Exit on error : invalid data.
		// popul.fill_with_missing(v::x);
		// popul.sum(v::x);

		// Using a lambda.
		assert(popul.sum(MSL_LMBD(true)) == popul.sum(v::one));
	}
	{
		std::cout << "Function member Population::weighted_sum\n";
		msl::Population popul(size);
		for (auto [i, ii] : msl::enumerate(indices))
			popul[ii].set_alive(i % 10).set_x(gnr()).set_weight(gnr());

		double total{};
		for (msl::RefConst ref : popul)
			if (ref.alive())
				total += ref.weight * ref.x;
		// On data member.
		assert(msl::approx_equal(popul.weighted_sum(v::x, v::weight), total));
		// Using a lambda.
		assert(msl::approx_equal(popul.weighted_sum(MSL_LMBD(ref.x), v::weight), total));
	}
	{
		std::cout << "Function member Population::sum_if\n";
		msl::Population popul(size);
		for (auto [i, ii] : msl::enumerate(indices))
			popul[ii].set_alive(i % 10).set_x(gnr()).set_female(gnr() % 2);

		double total{};
		for (msl::RefConst ref : popul)
			if (ref.alive() && ref.female())
				total += ref.x;
		// On data member and data member.
		assert(popul.sum_if(v::x, v::female) == total);
		// On data member and using a lambda.
		assert(popul.sum_if(v::x, MSL_LMBD(ref.female())) == total);
		// Using a lambda and on data member.
		assert(popul.sum_if(MSL_LMBD(ref.x), v::female) == total);
		// Using a lambda and a lambda.
		assert(popul.sum_if(MSL_LMBD(ref.x), v::female) == total);
	}
	{
		std::cout << "Function member Population::weighted_sum_if\n";
		msl::Population popul(size);
		for (auto [i, ii] : msl::enumerate(indices))
			popul[ii].set_alive(i % 10).set_x(gnr()).set_female(gnr() % 2).set_weight(gnr());

		double total{};
		for (msl::RefConst ref : popul)
			if (ref.alive() && ref.female())
				total += ref.weight * ref.x;

		// On data member and data member.
		assert(msl::approx_equal(popul.weighted_sum_if(v::x, v::female, v::weight), total));
		// On data member and using a lambda.
		assert(
		  msl::approx_equal(popul.weighted_sum_if(v::x, MSL_LMBD(ref.female()), v::weight), total));
		// Using a lambda and on data member.
		assert(msl::approx_equal(popul.weighted_sum_if(MSL_LMBD(ref.x), v::female, v::weight), total));
		// Using a lambda and a lambda.
		assert(msl::approx_equal(
		  popul.weighted_sum_if(MSL_LMBD(ref.x), MSL_LMBD(ref.female()), v::weight), total));
	}
	{
		std::cout << "Function member Population::merge_family - I\n";
		msl::Population popul(size);
		for (auto [i, ref] : msl::enumerate(popul))
			ref.set_alive(true)
			  .set_id_number(i)
			  .set_family_head(true)
			  .set_family_next(ref)
			  .set_household_head(true)
			  .set_household_next(ref)
			  .set_tax_fam_head(true)
			  .set_tax_fam_next(ref)
			  .set_ben_fam_head(true)
			  .set_ben_fam_next(ref);

		popul[0].merge_family(popul[1]).merge_family(popul[2]).merge_family(popul[3]).merge_family(
		  popul[4]);
		assert(popul[0].family_size() == 5);
		popul[2].death();
		assert(popul[0].family_size() == 4);
		popul[4].death();
		assert(popul[0].family_size() == 3);
		popul[0].death();
		assert(popul[1].family_size() == 2);
	}
	{
		std::cout << "Function member Population::merge_family - II\n";
		msl::Population popul(size);
		popul[0]
		  .set_alive(true)
		  .set_id_number(1)
		  .set_family_head(true)
		  .set_family_next(popul[0])
		  .set_household_head(true)
		  .set_household_next(popul[0])
		  .set_tax_fam_head(true)
		  .set_tax_fam_next(popul[0])
		  .set_ben_fam_head(true)
		  .set_ben_fam_next(popul[0]);
		for (size_t i{ 1 }; i < size / 2; ++i)
			popul.birth(popul[0]);
		auto n{ popul.count_if(v::one) };
		for (size_t i{}; i < 1000; ++i) {
			while (true) {
				auto const j{ gnr() % popul.size() };
				if (popul[j].alive()) {
					popul[j].death();
					break;
				}
			}
			size_t j;
			while (true) {
				j = gnr() % popul.size();
				if (popul[j].alive())
					break;
			}
			popul.birth(popul[j]);
		}
		assert(popul.count_if(v::one) == n);
	}
	{
		std::cout << "Function member CircIterAsForwardIter::sum\n";
		msl::Population popul(size);
		for (auto [i, ref] : msl::enumerate(popul))
			ref.set_alive(i % 10)
			  .set_x(gnr())
			  .set_female(gnr() % 2)
			  .set_family_head(true)
			  .set_family_next(ref);

		popul[1].merge_family(popul[2]).merge_family(popul[3]).merge_family(popul[4]);

		msl::Population const& const_pop{ popul };

		// When the individual is the head of the family.
		{
			double total{};
			for (msl::RefConst ref : const_pop[1].over_family())
				total += ref.x;
			assert(msl::approx_equal(total, const_pop[1].over_family().sum(v::x)));
		}
		// When the individual is not always the head of the family.
		{
			double total{};
			for (msl::RefConst ref : const_pop[1].over_family())
				total += ref.x;
			assert(msl::approx_equal(total, const_pop[1].family_next->over_family().sum(v::x)));
		}
		// Non const version.
		// When the individual is the head of the family.
		{
			double total{};
			for (msl::RefConst ref : popul[1].over_family())
				total += ref.x;
			assert(msl::approx_equal(total, popul[1].over_family().sum(v::x)));
		}
		// When the individual is not always the head of the family.
		{
			double total{};
			for (msl::RefConst ref : popul[1].over_family())
				total += ref.x;
			assert(msl::approx_equal(total, popul[1].family_next->over_family().sum(v::x)));
		}
	}
	{
		std::cout << "Function member Population::avg\n";
		msl::Population popul(size);
		for (auto [i, ii] : msl::enumerate(indices))
			popul[ii].set_alive(i % 10).set_std_living_10((i + 1) % 10).set_x(i % 3);

		// On a FLG data member.
		{
			assert(popul.avg(v::one) == 1);
		}

		// On a CTG data member.
		{
			assert(popul.avg(v::std_living_10) == popul.sum(v::std_living_10) / popul.count_if(v::one));
		}

		// On a FLT data member.
		{
			assert(popul.avg(v::x) == popul.sum(v::x) / popul.count_if(v::one));
		}

		// Throw an error : invalid variable.
		try {
			popul.avg(v::family_next);
			assert("not reached");
		} catch (...) {
		}
		// Exit on error : invalid data.
		// popul.fill_with_missing(v::x);
		// popul.avg(v::x);

		// Using a lambda.
		assert(popul.avg(MSL_LMBD(true)) == 1);
	}
	{
		std::cout << "Function member Population::weighted_avg\n";
		msl::Population popul(size);
		for (auto [i, ii] : msl::enumerate(indices))
			popul[ii].set_alive(i % 10).set_x(gnr()).set_weight(gnr());

		double total{};
		for (msl::RefConst ref : popul)
			if (ref.alive())
				total += ref.weight * ref.x;
		// On data member.
		assert(msl::approx_equal(popul.weighted_avg(v::x, v::weight), total / popul.sum(v::weight)));
		// Using a lambda.
		assert(msl::approx_equal(popul.weighted_avg(MSL_LMBD(ref.x), v::weight),
		                         total / popul.sum(v::weight)));
	}
	{
		std::cout << "Function member Population::avg_if\n";
		msl::Population popul(size);
		for (auto [i, ii] : msl::enumerate(indices))
			popul[ii].set_alive(i % 10).set_x(gnr()).set_female(gnr() % 2);

		double total{};
		for (msl::RefConst ref : popul)
			if (ref.alive() && ref.female())
				total += ref.x;
		// On data member and data member.
		assert(msl::approx_equal(popul.avg_if(v::x, v::female), total / popul.count_if(v::female)));
		// On data member and using a lambda.
		assert(msl::approx_equal(popul.avg_if(v::x, MSL_LMBD(ref.female())),
		                         total / popul.count_if(v::female)));
		// Using a lambda and on data member.
		assert(msl::approx_equal(popul.avg_if(MSL_LMBD(ref.x), v::female),
		                         total / popul.count_if(v::female)));
		// Using a lambda and a lambda.
		assert(msl::approx_equal(popul.avg_if(MSL_LMBD(ref.x), v::female),
		                         total / popul.count_if(v::female)));
	}
	{
		std::cout << "Function member Population::weighted_avg_if\n";
		msl::Population popul(size);
		for (auto [i, ii] : msl::enumerate(indices))
			popul[ii].set_alive(i % 10).set_x(gnr()).set_female(gnr() % 2).set_weight(gnr());

		double total{};
		for (msl::RefConst ref : popul)
			if (ref.alive() && ref.female())
				total += ref.weight * ref.x;

		// On data member and data member.
		assert(msl::approx_equal(popul.weighted_avg_if(v::x, v::female, v::weight),
		                         total / popul.sum_if(v::weight, v::female)));
		// On data member and using a lambda.
		assert(msl::approx_equal(popul.weighted_avg_if(v::x, MSL_LMBD(ref.female()), v::weight),
		                         total / popul.sum_if(v::weight, v::female)));
		// Using a lambda and on data member.
		assert(msl::approx_equal(popul.weighted_avg_if(MSL_LMBD(ref.x), v::female, v::weight),
		                         total / popul.sum_if(v::weight, v::female)));
		// Using a lambda and a lambda.
		assert(
		  msl::approx_equal(popul.weighted_avg_if(MSL_LMBD(ref.x), MSL_LMBD(ref.female()), v::weight),
		                    total / popul.sum_if(v::weight, v::female)));
	}
	{
		std::cout << "Function member CircIterAsForwardIter::avg\n";
		msl::Population popul(size);
		for (auto [i, ref] : msl::enumerate(popul))
			ref.set_alive(i % 10)
			  .set_x(gnr())
			  .set_female(gnr() % 2)
			  .set_family_head(true)
			  .set_family_next(ref);

		popul[1].merge_family(popul[2]).merge_family(popul[3]).merge_family(popul[4]);

		msl::Population const& const_pop{ popul };

		// When the individual is the head of the family.
		{
			double total{};
			for (msl::RefConst ref : const_pop[1].over_family())
				total += ref.x;
			assert(msl::approx_equal(total / const_pop[1].family_size(),
			                         const_pop[1].over_family().avg(v::x)));
		}
		// When the individual is not always the head of the family.
		{
			double total{};
			for (msl::RefConst ref : const_pop[1].over_family())
				total += ref.x;
			assert(msl::approx_equal(total / const_pop[1].family_size(),
			                         const_pop[1].family_next->over_family().avg(v::x)));
		}
		// Non const version.
		// When the individual is the head of the family.
		{
			double total{};
			for (msl::RefConst ref : popul[1].over_family())
				total += ref.x;
			assert(
			  msl::approx_equal(total / const_pop[1].family_size(), popul[1].over_family().avg(v::x)));
		}
		// When the individual is not always the head of the family.
		{
			double total{};
			for (msl::RefConst ref : popul[1].over_family())
				total += ref.x;
			assert(msl::approx_equal(total / const_pop[1].family_size(),
			                         popul[1].family_next->over_family().avg(v::x)));
		}
	}
	{
		std::cout << "Function member Population::std\n";
		msl::Population popul(size);
		for (auto [i, ii] : msl::enumerate(indices))
			popul[ii].set_alive(i % 10).set_std_living_10((i + 1) % 10).set_x(i % 3);

		// On a FLG data member.
		{
			assert(popul.std(v::one) == 0);
		}

		// On a CTG data member.
		{
			double s2{};
			double s{};
			size_t n{};
			for (msl::RefConst ref : popul)
				s2 += ref.alive() ? ref.std_living_10 * ref.std_living_10 : 0,
				  s += ref.alive() ? ref.std_living_10 : 0, n += ref.alive() ? 1 : 0;
			assert(msl::approx_equal(popul.std(v::std_living_10),
			                         std::sqrt(s2 / (n - 1) - (s / n) * (s / (n - 1)))));
		}

		// On a FLT data member.
		{
			double s2{};
			double s{};
			size_t n{};
			for (msl::RefConst ref : popul)
				s2 += ref.alive() ? ref.x * ref.x : 0, s += ref.alive() ? ref.x : 0,
				  n += ref.alive() ? 1 : 0;
			assert(msl::approx_equal(popul.std(v::x), std::sqrt(s2 / (n - 1) - (s / n) * (s / (n - 1)))));
		}

		// Throw an error : invalid variable.
		try {
			popul.std(v::family_next);
			assert("not reached");
		} catch (...) {
		}
		// Exit on error : invalid data.
		// popul.fill_with_missing(v::x);
		// popul.std(v::x);

		// Using a lambda.
		assert(popul.std(MSL_LMBD(true)) == 0);
	}
	{
		std::cout << "Function member Population::weighted_std\n";
		msl::Population popul(size);
		for (auto [i, ii] : msl::enumerate(indices))
			popul[ii].set_alive(i % 10).set_x(gnr()).set_weight(gnr());

		double s2{};
		double s{};
		size_t n{};
		double sw{};
		for (msl::RefConst ref : popul)
			if (ref.alive())
				s2 += ref.weight * ref.x * ref.x, s += ref.weight * ref.x, ++n, sw += ref.weight;
		// On data member.
		assert(msl::approx_equal(popul.weighted_std(v::x, v::weight),
		                         std::sqrt((s2 - s * (s / sw)) / (n - 1) * n / sw)));
		// Using a lambda.
		assert(msl::approx_equal(popul.weighted_std(MSL_LMBD(ref.x), v::weight),
		                         std::sqrt((s2 - s * (s / sw)) / (n - 1) * n / sw)));
	}
	{
		std::cout << "Function member Population::std_if\n";
		msl::Population popul(size);
		for (auto [i, ii] : msl::enumerate(indices))
			popul[ii].set_alive(i % 10).set_x(gnr()).set_female(gnr() % 2);

		double s2{};
		double s{};
		size_t n{};
		for (msl::RefConst ref : popul)
			if (ref.alive() && ref.female())
				s2 += ref.x * ref.x, s += ref.x, ++n;
		// On data member and data member.
		assert(msl::approx_equal(popul.std_if(v::x, v::female),
		                         std::sqrt(s2 / (n - 1) - (s / n) * (s / (n - 1)))));
		// On data member and using a lambda.
		assert(msl::approx_equal(popul.std_if(v::x, MSL_LMBD(ref.female())),
		                         std::sqrt(s2 / (n - 1) - (s / n) * (s / (n - 1)))));
		// Using a lambda and on data member.
		assert(msl::approx_equal(popul.std_if(MSL_LMBD(ref.x), v::female),
		                         std::sqrt(s2 / (n - 1) - (s / n) * (s / (n - 1)))));
		// Using a lambda and a lambda.
		assert(msl::approx_equal(popul.std_if(MSL_LMBD(ref.x), v::female),
		                         std::sqrt(s2 / (n - 1) - (s / n) * (s / (n - 1)))));
	}
	{
		std::cout << "Function member Population::weighted_std_if\n";
		msl::Population popul(size);
		for (auto [i, ii] : msl::enumerate(indices))
			popul[ii].set_alive(i % 10).set_x(gnr()).set_female(gnr() % 2).set_weight(gnr());

		double s2{};
		double s{};
		size_t n{};
		double sw{};
		for (msl::RefConst ref : popul)
			if (ref.alive() && ref.female())
				s2 += ref.weight * ref.x * ref.x, s += ref.weight * ref.x, ++n, sw += ref.weight;
		// On data member and data member.
		assert(msl::approx_equal(popul.weighted_std_if(v::x, v::female, v::weight),
		                         std::sqrt((s2 - s * (s / sw)) / (n - 1) * n / sw)));
		// On data member and using a lambda.
		assert(msl::approx_equal(popul.weighted_std_if(v::x, MSL_LMBD(ref.female()), v::weight),
		                         std::sqrt((s2 - s * (s / sw)) / (n - 1) * n / sw)));
		// Using a lambda and on data member.
		assert(msl::approx_equal(popul.weighted_std_if(MSL_LMBD(ref.x), v::female, v::weight),
		                         std::sqrt((s2 - s * (s / sw)) / (n - 1) * n / sw)));
		// Using a lambda and a lambda.
		assert(
		  msl::approx_equal(popul.weighted_std_if(MSL_LMBD(ref.x), MSL_LMBD(ref.female()), v::weight),
		                    std::sqrt((s2 - s * (s / sw)) / (n - 1) * n / sw)));
	}
	{
		std::cout << "Function member Population::min\n";
		msl::Population popul(size);
		for (auto [i, ii] : msl::enumerate(indices))
			popul[ii].set_alive(i % 10).set_std_living_10((i + 1) % 10).set_x(i % 3);

		// On a FLG data member.
		{
			assert(popul.min(v::one) == 1);
		}

		// On a CTG data member.
		{
			assert(popul.min(v::std_living_10) == 0);
		}

		// On a FLT data member.
		{
			auto min{ std::numeric_limits<double>::max() };
			for (msl::RefConst ref : popul)
				if (ref.alive() && (min > ref.x))
					min = ref.x;
			assert(popul.min(v::x) == min);
		}

		// Throw an error : invalid variable.
		try {
			popul.min(v::family_next);
			assert("not reached");
		} catch (...) {
		}
		// Exit on error : invalid data.
		// popul.fill_with_missing(v::x);
		// popul.min(v::x);

		// Using a lambda.
		assert(popul.min(MSL_LMBD(true)) == 1);
	}
	{
		std::cout << "Function member Population::min_if\n";
		msl::Population popul(size);
		for (auto [i, ii] : msl::enumerate(indices))
			popul[ii].set_alive(i % 10).set_x(gnr()).set_female(gnr() % 2);

		auto min{ std::numeric_limits<double>::max() };
		for (msl::RefConst ref : popul)
			if (ref.alive() && ref.female() && (min > ref.x))
				min = ref.x;
		// On data member and data member.
		assert(popul.min_if(v::x, v::female) == min);
		// On data member and using a lambda.
		assert(popul.min_if(v::x, MSL_LMBD(ref.female())) == min);
		// Using a lambda and on data member.
		assert(popul.min_if(MSL_LMBD(ref.x), v::female) == min);
		// Using a lambda and a lambda.
		assert(popul.min_if(MSL_LMBD(ref.x), v::female) == min);
	}
	{
		std::cout << "Function member Population::ref_min\n";
		msl::Population popul(size);
		for (auto [i, ii] : msl::enumerate(indices))
			popul[ii].set_alive(i % 10).set_female(true).set_std_living_10(10).set_x(gnr());

		// On a FLG data member.
		{
			popul[1].set_female(0);
			assert(popul.ref_min(v::female) == popul[1]);
			popul.ref_min(v::female).x = 0;
		}

		// On a CTG data member.
		{
			popul[1].set_std_living_10(0);
			assert(popul.ref_min(v::std_living_10) == popul[1]);
			popul.ref_min(v::std_living_10).x = 0;
		}

		// On a FLT data member and using a lambda.
		{
			auto min{ std::numeric_limits<double>::max() };
			msl::Individual const* ptr;
			for (msl::RefConst ref : popul)
				if (ref.alive() && (min > ref.x))
					min = ref.x, ptr = &ref;
			assert(popul.ref_min(v::x) == *ptr);
			assert(popul.ref_min(MSL_LMBD(ref.x)) == *ptr);
			popul.ref_min(v::x).x = 0;
			popul.ref_min(MSL_LMBD(ref.x)).x = 0;
		}

		// Throw an error : invalid variable.
		try {
			popul.ref_min(v::family_next);
			assert("not reached");
		} catch (...) {
		}
		// Exit on error : invalid data.
		// popul.fill_with_missing(v::x);
		// popul.ref_min(v::x);
	}
	{
		std::cout << "Function member Population::ref_min_if\n";
		msl::Population popul(size);
		for (auto [i, ii] : msl::enumerate(indices))
			popul[ii].set_alive(i % 10).set_x(gnr()).set_female(gnr() % 2);

		auto min{ std::numeric_limits<double>::max() };
		msl::Individual const* ptr;
		for (msl::RefConst ref : popul)
			if (ref.alive() && ref.female() && (min > ref.x))
				min = ref.x, ptr = &ref;
		// On data member and data member.
		assert(popul.ref_min_if(v::x, v::female) == *ptr);
		// On data member and using a lambda.
		assert(popul.ref_min_if(v::x, MSL_LMBD(ref.female())) == *ptr);
		// Using a lambda and on data member.
		assert(popul.ref_min_if(MSL_LMBD(ref.x), v::female) == *ptr);
		// Using a lambda and a lambda.
		assert(popul.ref_min_if(MSL_LMBD(ref.x), v::female) == *ptr);
		popul.ref_min_if(v::x, v::female).x = 0;
		popul.ref_min_if(v::x, MSL_LMBD(ref.female())).x = 0;
		popul.ref_min_if(MSL_LMBD(ref.x), v::female).x = 0;
		popul.ref_min_if(MSL_LMBD(ref.x), v::female).x = 0;
	}
	{
		std::cout << "Function member CircIterAsForwardIter::min\n";
		msl::Population popul(size);
		for (auto [i, ref] : msl::enumerate(popul))
			ref.set_alive(i % 10)
			  .set_x(gnr())
			  .set_female(gnr() % 2)
			  .set_family_head(true)
			  .set_family_next(ref);

		popul[1].merge_family(popul[2]).merge_family(popul[3]).merge_family(popul[4]);

		msl::Population const& const_pop{ popul };

		// When the individual is the head of the family.
		{
			double min{ std::numeric_limits<double>::max() };
			for (msl::RefConst ref : const_pop[1].over_family())
				if (min > ref.x)
					min = ref.x;
			assert(const_pop[1].over_family().min(v::x) == min);
			assert(const_pop[1].over_family().min(MSL_LMBD(ref.x)) == min);
		}
		// When the individual is not always the head of the family.
		{
			double min{ std::numeric_limits<double>::max() };
			for (msl::RefConst ref : const_pop[1].over_family())
				if (min > ref.x)
					min = ref.x;
			assert(const_pop[1].family_next->over_family().min(v::x) == min);
			assert(const_pop[1].family_next->over_family().min(MSL_LMBD(ref.x)) == min);
		}
		// Non const version.
		// When the individual is the head of the family.
		{
			double min{ std::numeric_limits<double>::max() };
			for (msl::RefConst ref : popul[1].over_family())
				if (min > ref.x)
					min = ref.x;
			assert(popul[1].over_family().min(v::x) == min);
			assert(popul[1].over_family().min(MSL_LMBD(ref.x)) == min);
		}
		// When the individual is not always the head of the family.
		{
			double min{ std::numeric_limits<double>::max() };
			for (msl::RefConst ref : popul[1].over_family())
				if (min > ref.x)
					min = ref.x;
			assert(popul[1].family_next->over_family().min(v::x) == min);
			assert(popul[1].family_next->over_family().min(MSL_LMBD(ref.x)) == min);
		}
	}
	{
		std::cout << "Function member CircIterAsForwardIter::ref_min\n";
		msl::Population popul(size);
		for (auto [i, ref] : msl::enumerate(popul))
			ref.set_alive(i % 10)
			  .set_x(gnr())
			  .set_female(gnr() % 2)
			  .set_family_head(true)
			  .set_family_next(ref);

		popul[1].merge_family(popul[2]).merge_family(popul[3]).merge_family(popul[4]);

		msl::Population const& const_pop{ popul };

		// When the individual is the head of the family.
		{
			double min{ std::numeric_limits<double>::max() };
			msl::Individual const* ptr;
			for (msl::RefConst ref : const_pop[1].over_family())
				if (min > ref.x)
					min = ref.x, ptr = &ref;
			assert(const_pop[1].over_family().ref_min(v::x) == *ptr);
			assert(const_pop[1].over_family().ref_min(MSL_LMBD(ref.x)) == *ptr);
		}
		// When the individual is not always the head of the family.
		{
			double min{ std::numeric_limits<double>::max() };
			msl::Individual const* ptr;
			for (msl::RefConst ref : const_pop[1].over_family())
				if (min > ref.x)
					min = ref.x, ptr = &ref;
			assert(const_pop[1].family_next->over_family().ref_min(v::x) == *ptr);
			assert(const_pop[1].family_next->over_family().ref_min(MSL_LMBD(ref.x)) == *ptr);
		}
		// Non const version.
		// When the individual is the head of the family.
		{
			double min{ std::numeric_limits<double>::max() };
			msl::Individual const* ptr;
			for (msl::RefConst ref : popul[1].over_family())
				if (min > ref.x)
					min = ref.x, ptr = &ref;
			assert(popul[1].over_family().ref_min(v::x) == *ptr);
			assert(popul[1].over_family().ref_min(MSL_LMBD(ref.x)) == *ptr);
			popul[1].over_family().ref_min(v::x).x = 0;
			popul[1].over_family().ref_min(MSL_LMBD(ref.x)).x = 0;
		}
		// When the individual is not always the head of the family.
		{
			double min{ std::numeric_limits<double>::max() };
			msl::Individual const* ptr;
			for (msl::RefConst ref : popul[1].over_family())
				if (min > ref.x)
					min = ref.x, ptr = &ref;
			assert(popul[1].family_next->over_family().ref_min(v::x) == *ptr);
			assert(popul[1].family_next->over_family().ref_min(MSL_LMBD(ref.x)) == *ptr);
			popul[1].family_next->over_family().ref_min(v::x).x = 0;
			popul[1].family_next->over_family().ref_min(MSL_LMBD(ref.x)).x = 0;
		}
	}
	{
		std::cout << "Function member Population::max\n";
		msl::Population popul(size);
		for (auto [i, ii] : msl::enumerate(indices))
			popul[ii].set_alive(i % 10).set_std_living_10((i + 1) % 10).set_x(i % 3);

		// On a FLG data member.
		{
			assert(popul.max(v::one) == 1);
		}

		// On a CTG data member.
		{
			assert(popul.max(v::std_living_10) == 9);
		}

		// On a FLT data member.
		{
			auto max{ -std::numeric_limits<double>::max() };
			for (msl::RefConst ref : popul)
				if (ref.alive() && (max < ref.x))
					max = ref.x;
			assert(popul.max(v::x) == max);
		}

		// Throw an error : invalid variable.
		try {
			popul.max(v::family_next);
			assert("not reached");
		} catch (...) {
		}
		// Exit on error : invalid data.
		// popul.fill_with_missing(v::x);
		// popul.max(v::x);

		// Using a lambda.
		assert(popul.max(MSL_LMBD(true)) == 1);
	}
	{
		std::cout << "Function member Population::max_if\n";
		msl::Population popul(size);
		for (auto [i, ii] : msl::enumerate(indices))
			popul[ii].set_alive(i % 10).set_x(gnr()).set_female(gnr() % 2);

		auto max{ -std::numeric_limits<double>::max() };
		for (msl::RefConst ref : popul)
			if (ref.alive() && ref.female() && (max < ref.x))
				max = ref.x;
		// On data member and data member.
		assert(popul.max_if(v::x, v::female) == max);
		// On data member and using a lambda.
		assert(popul.max_if(v::x, MSL_LMBD(ref.female())) == max);
		// Using a lambda and on data member.
		assert(popul.max_if(MSL_LMBD(ref.x), v::female) == max);
		// Using a lambda and a lambda.
		assert(popul.max_if(MSL_LMBD(ref.x), v::female) == max);
	}
	{
		std::cout << "Function member Population::ref_max\n";
		msl::Population popul(size);
		for (auto [i, ii] : msl::enumerate(indices))
			popul[ii].set_alive(i % 10).set_female(false).set_std_living_10(1).set_x(gnr());

		// On a FLG data member.
		{
			popul[1].set_female(1);
			assert(popul.ref_max(v::female) == popul[1]);
		}

		// On a CTG data member.
		{
			popul[1].set_std_living_10(10);
			popul.ref_max(v::std_living_10);
			assert(popul.ref_max(v::std_living_10) == popul[1]);
		}

		// On a FLT data member and using a lambda.
		{
			auto max{ -std::numeric_limits<double>::max() };
			msl::Individual const* ptr;
			for (msl::RefConst ref : popul)
				if (ref.alive() && (max < ref.x))
					max = ref.x, ptr = &ref;
			assert(popul.ref_max(v::x) == *ptr);
			assert(popul.ref_max(MSL_LMBD(ref.x)) == *ptr);
		}

		// Throw an error : invalid variable.
		try {
			popul.ref_max(v::family_next);
			assert("not reached");
		} catch (...) {
		}
		// Exit on error : invalid data.
		// popul.fill_with_missing(v::x);
		// popul.ref_max(v::x);
	}
	{
		std::cout << "Function member Population::ref_max_if\n";
		msl::Population popul(size);
		for (auto [i, ii] : msl::enumerate(indices))
			popul[ii].set_alive(i % 10).set_x(gnr()).set_female(gnr() % 2);

		auto ref_max{ -std::numeric_limits<double>::max() };
		msl::Individual const* ptr;
		for (msl::RefConst ref : popul)
			if (ref.alive() && ref.female() && (ref_max < ref.x))
				ref_max = ref.x, ptr = &ref;
		// On data member and data member.
		assert(popul.ref_max_if(v::x, v::female) == *ptr);
		// On data member and using a lambda.
		assert(popul.ref_max_if(v::x, MSL_LMBD(ref.female())) == *ptr);
		// Using a lambda and on data member.
		assert(popul.ref_max_if(MSL_LMBD(ref.x), v::female) == *ptr);
		// Using a lambda and a lambda.
		assert(popul.ref_max_if(MSL_LMBD(ref.x), v::female) == *ptr);
	}
	{
		std::cout << "Function member CircIterAsForwardIter::max\n";
		msl::Population popul(size);
		for (auto [i, ref] : msl::enumerate(popul))
			ref.set_alive(i % 10)
			  .set_x(gnr())
			  .set_female(gnr() % 2)
			  .set_family_head(true)
			  .set_family_next(ref);

		popul[1].merge_family(popul[2]).merge_family(popul[3]).merge_family(popul[4]);

		msl::Population const& const_pop{ popul };

		// When the individual is the head of the family.
		{
			double max{ -std::numeric_limits<double>::max() };
			for (msl::RefConst ref : const_pop[1].over_family())
				if (max < ref.x)
					max = ref.x;
			assert(const_pop[1].over_family().max(v::x) == max);
			assert(const_pop[1].over_family().max(MSL_LMBD(ref.x)) == max);
		}
		// When the individual is not always the head of the family.
		{
			double max{ -std::numeric_limits<double>::max() };
			for (msl::RefConst ref : const_pop[1].over_family())
				if (max < ref.x)
					max = ref.x;
			assert(const_pop[1].family_next->over_family().max(v::x) == max);
			assert(const_pop[1].family_next->over_family().max(MSL_LMBD(ref.x)) == max);
		}
		// Non const version.
		// When the individual is the head of the family.
		{
			double max{ -std::numeric_limits<double>::max() };
			for (msl::RefConst ref : popul[1].over_family())
				if (max < ref.x)
					max = ref.x;
			assert(popul[1].over_family().max(v::x) == max);
			assert(popul[1].over_family().max(MSL_LMBD(ref.x)) == max);
		}
		// When the individual is not always the head of the family.
		{
			double max{ -std::numeric_limits<double>::max() };
			for (msl::RefConst ref : popul[1].over_family())
				if (max < ref.x)
					max = ref.x;
			assert(popul[1].family_next->over_family().max(v::x) == max);
			assert(popul[1].family_next->over_family().max(MSL_LMBD(ref.x)) == max);
		}
	}
	{
		std::cout << "Function member CircIterAsForwardIter::ref_max\n";
		msl::Population popul(size);
		for (auto [i, ref] : msl::enumerate(popul))
			ref.set_alive(i % 10)
			  .set_x(gnr())
			  .set_female(gnr() % 2)
			  .set_family_head(true)
			  .set_family_next(ref);

		popul[1].merge_family(popul[2]).merge_family(popul[3]).merge_family(popul[4]);

		msl::Population const& const_pop{ popul };

		// When the individual is the head of the family.
		{
			double max{ -std::numeric_limits<double>::max() };
			msl::Individual const* ptr;
			for (msl::RefConst ref : const_pop[1].over_family())
				if (max < ref.x)
					max = ref.x, ptr = &ref;
			assert(const_pop[1].over_family().ref_max(v::x) == *ptr);
			assert(const_pop[1].over_family().ref_max(MSL_LMBD(ref.x)) == *ptr);
		}
		// When the individual is not always the head of the family.
		{
			double max{ -std::numeric_limits<double>::max() };
			msl::Individual const* ptr;
			for (msl::RefConst ref : const_pop[1].over_family())
				if (max < ref.x)
					max = ref.x, ptr = &ref;
			assert(const_pop[1].family_next->over_family().ref_max(v::x) == *ptr);
			assert(const_pop[1].family_next->over_family().ref_max(MSL_LMBD(ref.x)) == *ptr);
		}
		// Non const version.
		// When the individual is the head of the family.
		{
			double max{ -std::numeric_limits<double>::max() };
			msl::Individual const* ptr;
			for (msl::RefConst ref : popul[1].over_family())
				if (max < ref.x)
					max = ref.x, ptr = &ref;
			assert(popul[1].over_family().ref_max(v::x) == *ptr);
			assert(popul[1].over_family().ref_max(MSL_LMBD(ref.x)) == *ptr);
			popul[1].over_family().ref_max(v::x).x = 0;
			popul[1].over_family().ref_max(MSL_LMBD(ref.x)).x = 0;
		}
		// When the individual is not always the head of the family.
		{
			double max{ -std::numeric_limits<double>::max() };
			msl::Individual const* ptr;
			for (msl::RefConst ref : popul[1].over_family())
				if (max < ref.x)
					max = ref.x, ptr = &ref;
			assert(popul[1].family_next->over_family().ref_max(v::x) == *ptr);
			assert(popul[1].family_next->over_family().ref_max(MSL_LMBD(ref.x)) == *ptr);
			popul[1].family_next->over_family().ref_max(v::x).x = 0;
			popul[1].family_next->over_family().ref_max(MSL_LMBD(ref.x)).x = 0;
		}
	}
	{
		std::cout << "Function member Population::pctl\n";
		for (size_t n : { 10000, 10001 }) {
			msl::Population popul(n);
			for (size_t step : { 1, 2 }) {
				for (auto [i, individual] : msl::enumerate(popul))
					individual.set_alive(true).set_x(step * i + 1);
				auto const pctl{ popul.pctl(v::x) };
				// The difference between the first percentile and the min is equal to the difference
				// between the max and the last percentile.
				assert(msl::approx_equal(pctl[1] - pctl[0], pctl[100] - pctl[99]));
				// Percentile are uniformly distributed.
				assert(msl::approx_equal(pctl[10] - pctl[9], pctl[11] - pctl[10]));
			}
		}
	}
	{
		std::cout << "Function member Population::weighted_pctl\n";
		for (size_t n : { 10000, 10001 }) {
			msl::Population popul(n);
			for (size_t step : { 1, 2 }) {
				for (auto [i, individual] : msl::enumerate(popul))
					individual.set_alive(true).set_x(step * i + 1);
				for (size_t weight : { 1, 2 }) {
					popul.fill(v::weight, weight);
					auto const pctl{ popul.pctl(v::x) };
					auto const w_pctl{ popul.weighted_pctl(v::x, v::weight) };
					// The weighted percentiles are equal to the unweighted percentiles.
					for (size_t i : { 0, 1, 50, 99, 100 })
						assert(pctl[i] == w_pctl[i]);
				}
			}
		}
	}
	{
		std::cout << "Function member Population::gini\n";
		msl::Population popul(1000);
		for (auto [i, individual] : msl::enumerate(popul))
			individual.set_alive(true).set_x(i + 1);
		{
			auto const gini{ popul.gini(v::x) };
			assert(msl::approx_equal(gini, .333));
		}
		for (msl::Ref individual : popul)
			individual.set_x(0);
		popul[0].set_x(100);
		{
			auto const gini{ popul.gini(v::x) };
			assert(msl::approx_equal(gini, (1000. - 1) / 1000));
		}
		for (msl::Ref individual : popul)
			individual.set_x(100);
		{
			auto const gini{ popul.gini(v::x) };
			assert(gini == 0);
		}
	}
	{
		std::cout << "Function member Population::weighted_gini\n";
		{
			msl::Population popul(1000);
			for (auto [i, individual] : msl::enumerate(popul))
				individual.set_alive(true).set_x(i + 1).set_weight(1);
			{
				auto const gini{ popul.weighted_gini(v::x, v::weight) };
				assert(msl::approx_equal(gini, .333));
			}
			for (msl::Ref individual : popul)
				individual.set_x(0);
			popul[0].set_x(100);
			{
				auto const gini{ popul.weighted_gini(v::x, v::weight) };
				assert(msl::approx_equal(gini, (1000. - 1) / 1000));
			}
			for (auto [i, individual] : msl::enumerate(popul))
				individual.set_x(1);
			{
				auto const gini{ popul.weighted_gini(v::x, v::weight) };
				assert(gini == 0);
			}
		}
	}
	{
		std::cout << "Function member Population::sum_by\n";
		{
			msl::Population popul(size);
			for (auto [i, individual] : msl::enumerate(popul))
				individual.set_alive(true).set_x(i + 1).set_std_living_10(i % 10);
			{
				auto const sum_by{ popul.sum_by(v::x, v::std_living_10) };
				assert(sum_by.find(0)->second.second ==
				       popul.sum_if(v::x, [](msl::RefConst ref) { return ref.std_living_10 == 0; }));
			}
			{
				auto const sum_by{ popul.sum_by([](msl::RefConst ref) { return ref.x; },
					                              v::std_living_10) };
				assert(sum_by.find(0)->second.second ==
				       popul.sum_if(v::x, [](msl::RefConst ref) { return ref.std_living_10 == 0; }));
			}
		}
	}
	{
		std::cout << "Function member Population::weighted_sum_by\n";
		{
			msl::Population popul(size);
			for (auto [i, individual] : msl::enumerate(popul))
				individual.set_alive(true).set_x(i + 1).set_std_living_10(i % 10).set_weight(i % 3 + 10);
			{
				auto const weighted_sum_by{ popul.weighted_sum_by(v::x, v::std_living_10, v::weight) };
				assert(weighted_sum_by.find(0)->second.second ==
				       popul.weighted_sum_if(
				         v::x, [](msl::RefConst ref) { return ref.std_living_10 == 0; }, v::weight));
			}
			{
				auto const weighted_sum_by{ popul.weighted_sum_by(
					[](msl::RefConst ref) { return ref.x; }, v::std_living_10, v::weight) };
				assert(weighted_sum_by.find(0)->second.second ==
				       popul.weighted_sum_if(
				         v::x, [](msl::RefConst ref) { return ref.std_living_10 == 0; }, v::weight));
			}
		}
	}
	{
		std::cout << "Function member Population::sum_if_by\n";
		{
			msl::Population popul(size);
			for (auto [i, individual] : msl::enumerate(popul))
				individual.set_alive(true).set_x(i + 1).set_std_living_10(i % 10).set_female(i % 3);
			{
				auto const sum_if_by{ popul.sum_if_by(v::x, v::female, v::std_living_10) };
				assert(sum_if_by.find(0)->second.second == popul.sum_if(v::x, [](msl::RefConst ref) {
					return ref.female() && (ref.std_living_10 == 0);
				}));
			}
			{
				auto const sum_if_by{ popul.sum_if_by([](msl::RefConst ref) { return ref.x; },
					                                    [](msl::RefConst ref) { return ref.female(); },
					                                    v::std_living_10) };
				assert(sum_if_by.find(0)->second.second == popul.sum_if(v::x, [](msl::RefConst ref) {
					return ref.female() && (ref.std_living_10 == 0);
				}));
			}
			{
				auto const sum_if_by{ popul.sum_if_by(
					[](msl::RefConst ref) { return ref.x; }, v::female, v::std_living_10) };
				assert(sum_if_by.find(0)->second.second == popul.sum_if(v::x, [](msl::RefConst ref) {
					return ref.female() && (ref.std_living_10 == 0);
				}));
			}
			{
				auto const sum_if_by{ popul.sum_if_by(
					v::x, [](msl::RefConst ref) { return ref.female(); }, v::std_living_10) };
				assert(sum_if_by.find(0)->second.second == popul.sum_if(v::x, [](msl::RefConst ref) {
					return ref.female() && (ref.std_living_10 == 0);
				}));
			}
		}
	}
	{
		std::cout << "Function member Population::weighted_sum_if_by\n";
		{
			msl::Population popul(size);
			for (auto [i, individual] : msl::enumerate(popul))
				individual.set_alive(true)
				  .set_x(i + 1)
				  .set_std_living_10(i % 10)
				  .set_female(i % 3)
				  .set_weight(i % 5 + 10);
			{
				auto const weighted_sum_if_by{ popul.weighted_sum_if_by(
					v::x, v::female, v::std_living_10, v::weight) };
				assert(weighted_sum_if_by.find(0)->second.second ==
				       popul.weighted_sum_if(
				         v::x,
				         [](msl::RefConst ref) { return ref.female() && (ref.std_living_10 == 0); },
				         v::weight));
			}
			{
				auto const weighted_sum_if_by{ popul.weighted_sum_if_by(
					[](msl::RefConst ref) { return ref.x; },
					[](msl::RefConst ref) { return ref.female(); },
					v::std_living_10,
					v::weight) };
				assert(weighted_sum_if_by.find(0)->second.second ==
				       popul.weighted_sum_if(
				         v::x,
				         [](msl::RefConst ref) { return ref.female() && (ref.std_living_10 == 0); },
				         v::weight));
			}
			{
				auto const weighted_sum_if_by{ popul.weighted_sum_if_by(
					[](msl::RefConst ref) { return ref.x; }, v::female, v::std_living_10, v::weight) };
				assert(weighted_sum_if_by.find(0)->second.second ==
				       popul.weighted_sum_if(
				         v::x,
				         [](msl::RefConst ref) { return ref.female() && (ref.std_living_10 == 0); },
				         v::weight));
			}
			{
				auto const weighted_sum_if_by{ popul.weighted_sum_if_by(
					v::x, [](msl::RefConst ref) { return ref.female(); }, v::std_living_10, v::weight) };
				assert(weighted_sum_if_by.find(0)->second.second ==
				       popul.weighted_sum_if(
				         v::x,
				         [](msl::RefConst ref) { return ref.female() && (ref.std_living_10 == 0); },
				         v::weight));
			}
		}
	}
	{
		std::cout << "Function member Population::align\n";
		{
			msl::Population popul(size);
			for (auto [i, ii] : msl::enumerate(indices))
				popul[ii].set_alive(i % 10).set_f01((i + 1) % 2).set_prob(0);
			popul.for_each_if(MSL_FNCTN(ref.prob = .2), v::f01);
			auto const count{ size_t(popul.sum(v::prob) + .5) };
			assert(popul.align(count, v::prob, MSL_FNCTN(ref.set_f01(true)), true) == count);
		}
		msl::Population popul(size);
		for (msl::Ref ref : popul)
			ref.set_alive(true).set_prob(.01);
		popul.fill_if_with_missing(v::prob, MSL_LMBD(popul.get_index(ref) == 1));
		assert(popul.align(1, v::prob, MSL_FNCTN(ref.set_f01(true))) == 0);
		popul.fill(v::prob, 0);
		assert(popul.align(1, v::prob, MSL_FNCTN(ref.set_f01(true))) == 0);
	}
	{
		std::cout << "Variable to_day\n";
		msl::to_day = msl::days_from_civil(2023, 1, 1);
		msl::Population popul(size);
		for (msl::Ref ref : popul)
			ref.set_alive(true).set_birthday_ymd(2023, 1, 1);
		assert(popul[0].age() == 0);
	}
	{
		std::cout << "Function member Individual::has_partner\n";
		{
			msl::Population popul(size);
			popul[0].set_partner(popul[1]);
			assert(popul[0].has_partner() == true);
		}
	}
	std::cout << "Normal exit of the test program.\n" << std::endl;
	return 0;
}
